from collections import defaultdict
from operator import itemgetter
import re
from typing import Dict, Iterable, List, Match, Optional, Tuple, Union

from .errors import MarkupError
from .style import Style
from .text import Span, Text
from ._emoji_replace import _emoji_replace


re_tags = re.compile(r"(\[\[)|(\]\])|(\[.*?\])")


def escape(markup: str) -> str:
    """Escapes text so that it won't be interpreted as markup. 

    Args:
        markup (str): Content to be inserted in to markup.

    Returns:
        str: Markup with square brackets escaped.
    """
    return markup.replace("[", "[[").replace("]", "]]")


def _parse(markup: str) -> Iterable[Tuple[Optional[str], Optional[str]]]:
    """Parse markup in to an iterable of pairs of text, tag.
    
    Args:
        markup (str): A string containing console markup
    
    """
    position = 0
    for match in re_tags.finditer(markup):
        escape_open, escape_close, tag_text = match.groups()
        start, end = match.span()
        if start > position:
            yield markup[position:start], None
        if tag_text is not None:
            yield None, tag_text
        else:
            yield (escape_open and "[") or (escape_close and "]"), None  # type: ignore
        position = end
    if position < len(markup):
        yield markup[position:], None


def render(markup: str, style: Union[str, Style] = "", emoji: bool = True) -> Text:
    """Render console markup in to a Text instance.

    Args:
        markup (str): A string containing console markup.
        emoji (bool, optional): Also render emoji code. Defaults to True.
    
    Raises:
        MarkupError: If there is a syntax error in the markup.
    
    Returns:
        Text: A test instance.
    """
    text = Text(style=style)
    append = text.append

    styles: Dict[str, List[int]] = defaultdict(list)
    style_stack: List[str] = []
    normalize = Style.normalize
    emoji_replace = _emoji_replace

    spans: List[Span] = []
    append_span = spans.append

    _Span = Span

    for plain_text, tag in _parse(markup):
        if plain_text is not None:
            append(emoji_replace(plain_text) if emoji else plain_text)
        if tag is not None:
            if tag.startswith("[/"):  # Closing tag
                style_name = tag[2:-1].strip()
                if style_name:  # explicit close
                    style_name = normalize(style_name)
                else:  # implicit close
                    try:
                        style_name = style_stack[-1]
                    except IndexError:
                        raise MarkupError(
                            f"closing tag '[/]' at position {len(text)} has nothing to close"
                        )
                try:
                    style_position = styles[style_name].pop()
                except (KeyError, IndexError):
                    raise MarkupError(
                        f"closing tag {tag!r} at position {len(text)} doesn't match open tag"
                    )
                style_stack.remove(style_name)
                append_span(_Span(style_position, len(text), style_name))
            else:  # Opening tag
                style_name = normalize(tag[1:-1].strip())
                styles[style_name].append(len(text))
                style_stack.append(style_name)

    text_length = len(text)
    while style_stack:
        style_name = style_stack.pop()
        style_position = styles[style_name].pop()
        append_span(_Span(style_position, text_length, style_name))

    text.spans = sorted(spans)

    return text


if __name__ == "__main__":  # pragma: no cover
    from rich import print
    from rich.console import Console

    console = Console(highlight=False)

    console.print("[bold]1 [not bold]2[/] 3[/]")

    console.print("[green]XXX[blue]XXX[/]XXX[/]")



from typing import List


def ratio_divide(
    total: int, ratios: List[int], minimums: List[int] = None
) -> List[int]:
    """Divide an integer total in to parts based on ratios.
    
    Args:
        total (int): The total to divide.
        ratios (List[int]): A list of integer rations.
        minimums (List[int]): List of minimum values for each slot. 
    
    Returns:
        List[int]: A list of integers garanteed to sum to total.
    """
    total_ratio = sum(ratios)
    assert total_ratio > 0, "Sum of ratios must be > 0"

    total_remaining = total
    distributed_total: List[int] = []
    append = distributed_total.append
    if minimums is None:
        _minimums = [0] * len(ratios)
    else:
        _minimums = minimums
    for ratio, minimum in zip(ratios, _minimums):
        if total_ratio > 0:
            distributed = max(minimum, round(ratio * total_remaining / total_ratio))
        else:
            distributed = total_remaining
        append(distributed)
        total_ratio -= ratio
        total_remaining -= distributed
    return distributed_total



import io
from typing import Any, IO, Union

from .console import Console as BaseConsole
from .style import Style

JUPYTER_HTML_FORMAT = """\
<pre style="white-space:pre;overflow-x:auto;line-height:1em;font-family:Menlo,'DejaVu Sans Mono',consolas,'Courier New',monospace">{code}</pre>
"""


class JupyterRenderable:
    """A shim to write html to Jupyter notebook."""

    def __init__(self, text: str, html: str) -> None:
        self.text = text
        self.html = html

    def __str__(self) -> str:
        return self.text

    def _repr_html_(self) -> str:
        return self.html


class Console(BaseConsole):
    def __init__(self, **kwargs) -> None:
        kwargs["file"] = io.StringIO()
        kwargs["record"] = True
        if "width" not in kwargs:
            kwargs["width"] = 100
        super().__init__(**kwargs)

    def is_terminal(self) -> bool:
        return True

    def print(  # type: ignore
        self,
        *objects: Any,
        sep=" ",
        end="\n",
        file: IO[str] = None,
        style: Union[str, Style] = None,
        emoji: bool = None,
        markup: bool = None,
        highlight: bool = None,
        flush: bool = False
    ) -> JupyterRenderable:
        r"""Print to the console.

        Args:
            objects (positional args): Objects to log to the terminal.
            sep (str, optional): String to write between print data. Defaults to " ".
            end (str, optional): String to write at end of print data. Defaults to "\n".
            style (Union[str, Style], optional): A style to apply to output. Defaults to None.
            emoji (Optional[bool], optional): Enable emoji code, or ``None`` to use console default. Defaults to None.
            markup (Optional[bool], optional): Enable markup, or ``None`` to use console default. Defaults to None
            highlight (Optional[bool], optional): Enable automatic highlighting, or ``None`` to use console default. Defaults to None.
        """
        if file is None:
            super().print(
                *objects,
                sep=sep,
                end=end,
                style=style,
                emoji=emoji,
                markup=markup,
                highlight=highlight,
            )
        else:
            Console(file=file).print(
                *objects,
                sep=sep,
                end=end,
                style=style,
                emoji=emoji,
                markup=markup,
                highlight=highlight,
            )

        html = self.export_html(code_format=JUPYTER_HTML_FORMAT, inline_styles=True)
        text = self.file.getvalue()  # type: ignore
        self.file = io.StringIO()
        jupyter_renderable = JupyterRenderable(text, html)
        return jupyter_renderable


console = Console()
print = console.print
